/*
  circular queue data structure for network i/o buffering
  which suitable for windows iocp / linux-epoll socket model.
  copyright(c) 1990 - 2011 krkim all rights reserved.
  http://krkim.net

 */

#ifdef _WINDOWS
#include "stdafx.h"
#include <windows.h>
#include <tchar.h>
#include <assert.h>
#else /*linux or unix g++*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>    /* isdegit() */
#include <assert.h>
#endif
#include "iocq.h"

int CIOCQ::m_iocq_count = 0;
CIOCQ::CIOCQ()
{
}

CQUEUE * CIOCQ::Init(int buffsize)
{
	CQUEUE *pIOCQ = new CQUEUE;
	if(!pIOCQ) return NULL;
	m_iocq_count++;

	pIOCQ->buff = new char[buffsize+1];
	memset(pIOCQ->buff, 0, buffsize+1);
	pIOCQ->front = pIOCQ->rear = 0;
	pIOCQ->buffsize = buffsize;
	pIOCQ->datasize = 0;
	return pIOCQ;
}

void CIOCQ::Free(CQUEUE *pIOCQ)
{
	if(pIOCQ->buff)
		delete pIOCQ->buff;
	pIOCQ->buff = NULL;
	delete pIOCQ;
	m_iocq_count--;
}

//----------------------------------------------------------------------------
// ť  ʱȭ
//----------------------------------------------------------------------------
void CIOCQ::Clear(CQUEUE *pIOCQ)
{
	pIOCQ->front = pIOCQ->rear = 0;
	pIOCQ->datasize = 0;
}

//----------------------------------------------------------------------------
//    (  ó) Move Rear Pointer
//----------------------------------------------------------------------------
int	CIOCQ::ShiftRear(CQUEUE *pIOCQ,int nShift)
{
	pIOCQ->rear = ( pIOCQ->rear + nShift ) % pIOCQ->buffsize;
	pIOCQ->datasize += nShift;
	if(pIOCQ->datasize > pIOCQ->buffsize)
		pIOCQ->datasize = pIOCQ->buffsize;
	return pIOCQ->rear;
}

//----------------------------------------------------------------------------
// б   (  ó)
//----------------------------------------------------------------------------
int	CIOCQ::ShiftFront(CQUEUE *pIOCQ,int nShift)
{
	pIOCQ->front = (pIOCQ->front + nShift) % pIOCQ->buffsize;
	pIOCQ->datasize -= nShift;
	if(pIOCQ->datasize <= 0){
		pIOCQ->front = pIOCQ->rear = 0;
		pIOCQ->datasize = 0;
	}
	return pIOCQ->front;
}

bool CIOCQ::IsFull(CQUEUE *pIOCQ)
{
	if(pIOCQ->datasize == pIOCQ->buffsize)
		return true;
	return false;
}

//----------------------------------------------------------------------------
// Ÿ  ˻
//----------------------------------------------------------------------------
bool CIOCQ::IsEmpty(CQUEUE *pIOCQ)
{
	if(pIOCQ->datasize == 0)
		return true;
	return false;
}

//----------------------------------------------------------------------------
// ť ⺻Լ
//----------------------------------------------------------------------------
bool CIOCQ::enqueue(CQUEUE *pIOCQ,char ch)
{
	if(IsFull(pIOCQ))
		return false;
	*(pIOCQ->buff + pIOCQ->rear) = ch;
	ShiftRear(pIOCQ,1);
	return true;
}

//----------------------------------------------------------------------------
// ť ⺻Լ
//----------------------------------------------------------------------------
bool CIOCQ::dequeue(CQUEUE *pIOCQ,char *ch)
{
	*ch = 0;
	if(IsEmpty(pIOCQ))
		return false;
	*ch = *(pIOCQ->buff + pIOCQ->front);
	ShiftFront(pIOCQ,1);
	return true;
}

//----------------------------------------------------------------------------
// ۳ offset Ÿ 
//----------------------------------------------------------------------------
char * CIOCQ::getoffsetptr(CQUEUE *pIOCQ,int nOffset)
{
	nOffset = nOffset % pIOCQ->buffsize;
	return (pIOCQ->buff + nOffset);
}

//----------------------------------------------------------------------------
//nOffset  ۳ sequential ϰ Է° ũ
//----------------------------------------------------------------------------
int CIOCQ::getoffsetsize(CQUEUE *pIOCQ,int nOffset)
{
	int freesize;
	nOffset = nOffset % pIOCQ->buffsize;
	if(nOffset < pIOCQ->front)
		freesize = (pIOCQ->front - nOffset);
	else{
		freesize = pIOCQ->buffsize - nOffset;
	}
	return freesize;
}

//----------------------------------------------------------------------------
//Get Data size
//----------------------------------------------------------------------------
int	CIOCQ::GetDataSize(CQUEUE *pIOCQ)
{
	return pIOCQ->datasize;
}

//----------------------------------------------------------------------------
//Get Free Queue size
//----------------------------------------------------------------------------
int	CIOCQ::GetAvailSize(CQUEUE *pIOCQ)
{
	return pIOCQ->buffsize - pIOCQ->datasize;
}

//----------------------------------------------------------------------------
//Get Data Buffer From Front Pointer
//lpRestDataLen :frontκ ۳ Ÿũ
// |-->F         R-->| , R - F size
// |-->R         F-->| , F--> size only(except -->R)
//----------------------------------------------------------------------------
char * CIOCQ::GetFrontPtr(CQUEUE *pIOCQ)
{
	return pIOCQ->buff + pIOCQ->front;
}

//----------------------------------------------------------------------------
//Get Rear Pointer to Write Data
//----------------------------------------------------------------------------
char * CIOCQ::GetRearPtr(CQUEUE *pIOCQ)
{
	return (pIOCQ->buff + pIOCQ->rear);
}

//----------------------------------------------------------------------------
//return the size that add to sequentially new data 
//from current rear to end of buffer.
//----------------------------------------------------------------------------
int CIOCQ::GetRearSize(CQUEUE *pIOCQ)
{
	return getoffsetsize(pIOCQ,pIOCQ->rear);
}

//----------------------------------------------------------------------------
//Ǵ  ȰϿ  Է ִ  ȮѴ
//,front(F)  rear(R)   , front óκ
// ִ ŭ Ǵ rear  Է° ȮѴ.
//IOCP  ſû ۿ GetRearPtr ֱ Ͽ, Ȯ.
//ֳϸ,IOCP۴ circular  ƴ Ϲ linear-buffer(array).
// |-->F         R-->| , ϶ F 翩 ŭ R  ȴ.
// |F          R---->| ǵ Ѵ.
// rear  front   ʿ,Է° (rear~front)
// ̾   ̴.
//ϰ:Է  (rear ) Ÿ ũ
//IOCP  ۿ  ũ⸦ Ѱٶ Ѵ.
//ShiftData() ȣ ũ GetRearPtr()  
//ex)
// ޱ IOCP û,
//۸ ϸ鼭 sequential  ũ⸦ ´.
//int nAvailBuffSize = CIOCQ::ShiftData(pIOBuff);
//char *buf = CIOCQ::GetRearPtr(pIOBuff);
//----------------------------------------------------------------------------

int CIOCQ::ShiftData(CQUEUE *pIOCQ)
{
	if(pIOCQ->rear >= pIOCQ->front && pIOCQ->front > 0){
		for(int i = 0; i < pIOCQ->datasize; i++){
			*(pIOCQ->buff + i) = *(pIOCQ->buff + pIOCQ->front + i);
		}
		pIOCQ->rear -= pIOCQ->front;
		pIOCQ->front = 0;
	}
	return GetRearSize(pIOCQ);
}

//----------------------------------------------------------------------------
//bAll : TRUE = nLen ŭ θ Էϱ Ҷ
//       FALSE =  ŭ Էϰ Է° 
//return : Էµ Ÿ 
//----------------------------------------------------------------------------
int CIOCQ::Enqueue(CQUEUE *pIOCQ,char *pData, int nLen,bool bAll)
{
	int nDataSize = GetDataSize(pIOCQ);
	int nAvailSize = GetAvailSize(pIOCQ);
	int nWrite = 0;
	if(bAll && nAvailSize < nLen)//絥Ÿ Է ϴ 
		return 0;
	for(nWrite = 0; nWrite < nLen;nWrite++){
		if(!enqueue(pIOCQ,*(pData+nWrite)))
			break;
	}
	return nWrite;
}

//----------------------------------------------------------------------------
//Peek Data from Front Pointer ( shift front  ʴ´)
//nLen:  ũ
//nOffset : front κ , ۳ ġ 
//          0 =   ġ front κ 
//ϰ:   ũ(0 = empty)
//----------------------------------------------------------------------------
int CIOCQ::Peek(CQUEUE *pIOCQ,char *pDest, int nLen,int nOffset)
{
	int nRead = 0;
	int nDataSize = GetDataSize(pIOCQ);
	int nSkipSize = nOffset;
	int nToLen = (nDataSize < nOffset + nLen) ? nDataSize - nOffset: nLen;
	int front = (pIOCQ->front + nOffset) % pIOCQ->buffsize;//ť ȯ
	for(nRead = 0; nRead < nToLen;nRead++){
		char ch = *(pIOCQ->buff + front);
		*(pDest+nRead) = ch;
		front = (front + 1) % pIOCQ->buffsize;
	}
	*(pDest+nRead) = 0;
	return nRead;
}

//----------------------------------------------------------------------------
//GetData from Front Pointer and shift front
//nOffset : front κ ,, ۳ 
//nLen: ϴ Ÿ ũ
//return:  Ÿ ũ
//----------------------------------------------------------------------------
int CIOCQ::Dequeue(CQUEUE *pIOCQ,char *pDest, int nLen,int nOffset)
{
	int nRead = 0;
	nRead = Peek(pIOCQ,pDest,nLen,nOffset);
	ShiftFront(pIOCQ,nOffset + nRead);
	return nRead;
}
